from cognite.client.data_classes.data_modeling import filters
from cognite.client.data_classes.data_modeling.containers import (
    ConstraintIdentifier,
    Container,
    ContainerApply,
    ContainerApplyList,
    ContainerDirectRelation,
    ContainerFilter,
    ContainerList,
    ContainerPropertyIdentifier,
    IndexIdentifier,
    RequiresConstraintDefinition,
    UniquenessConstraintDefinition,
)
from cognite.client.data_classes.data_modeling.data_models import (
    DataModel,
    DataModelApply,
    DataModelApplyList,
    DataModelFilter,
    DataModelList,
    DataModelsSort,
)
from cognite.client.data_classes.data_modeling.data_types import (
    Boolean,
    CDFExternalIdReference,
    Date,
    DirectRelation,
    DirectRelationReference,
    FileReference,
    Float32,
    Float64,
    Int32,
    Int64,
    Json,
    Primitive,
    PropertyType,
    SequenceReference,
    Text,
    TimeSeriesReference,
    Timestamp,
)
from cognite.client.data_classes.data_modeling.filters import Filter
from cognite.client.data_classes.data_modeling.ids import (
    ContainerId,
    ContainerIdentifier,
    DataModelId,
    DataModelIdentifier,
    DataModelingId,
    VersionedDataModelingId,
    ViewId,
    ViewIdentifier,
)
from cognite.client.data_classes.data_modeling.spaces import Space, SpaceApply, SpaceApplyList, SpaceList
from cognite.client.data_classes.data_modeling.views import (
    ConnectionDefinition,
    MappedApplyPropertyDefinition,
    MappedPropertyDefinition,
    SingleHopConnectionDefinition,
    View,
    ViewApply,
    ViewApplyList,
    ViewFilter,
    ViewList,
    ViewPropertyDefinition,
)

__all__ = [
    "ViewIdentifier",
    "ViewApply",
    "ViewApplyList",
    "MappedApplyPropertyDefinition",
    "VersionedDataModelingId",
    "DataModelingId",
    "ContainerIdentifier",
    "DataModelIdentifier",
    "filters",
    "DirectRelation",
    "Filter",
    "DirectRelationReference",
    "DataModel",
    "DataModelList",
    "DataModelsSort",
    "DataModelFilter",
    "DataModelApply",
    "DataModelApplyList",
    "ContainerFilter",
    "ViewFilter",
    "MappedPropertyDefinition",
    "ConnectionDefinition",
    "SingleHopConnectionDefinition",
    "ViewPropertyDefinition",
    "Space",
    "SpaceList",
    "SpaceApply",
    "SpaceApplyList",
    "View",
    "ViewList",
    "Container",
    "ContainerList",
    "ContainerApply",
    "ContainerApplyList",
    "IndexIdentifier",
    "ConstraintIdentifier",
    "ContainerPropertyIdentifier",
    "Primitive",
    "CDFExternalIdReference",
    "ContainerDirectRelation",
    "RequiresConstraintDefinition",
    "UniquenessConstraintDefinition",
    "ContainerId",
    "ViewId",
    "DataModelId",
    "Text",
    "Boolean",
    "Float32",
    "Float64",
    "Int32",
    "Int64",
    "Timestamp",
    "Date",
    "Json",
    "TimeSeriesReference",
    "FileReference",
    "SequenceReference",
    "PropertyType",
]



from __future__ import annotations

from typing import Iterator, Sequence, cast, overload

from cognite.client._api_client import APIClient
from cognite.client._constants import LIST_LIMIT_DEFAULT
from cognite.client.data_classes.data_modeling.spaces import Space, SpaceApply, SpaceList
from cognite.client.utils._identifier import DataModelingIdentifierSequence


class SpacesAPI(APIClient):
    _RESOURCE_PATH = "/models/spaces"

    @overload
    def __call__(
        self,
        chunk_size: None = None,
        limit: int = None,
    ) -> Iterator[Space]:
        ...

    @overload
    def __call__(
        self,
        chunk_size: int,
        limit: int = None,
    ) -> Iterator[SpaceList]:
        ...

    def __call__(
        self,
        chunk_size: int = None,
        limit: int = None,
    ) -> Iterator[Space] | Iterator[SpaceList]:
        """Iterate over spaces

        Fetches spaces as they are iterated over, so you keep a limited number of spaces in memory.

        Args:
            chunk_size (int, optional): Number of spaces to return in each chunk. Defaults to yielding one space a time.
            limit (int, optional): Maximum number of spaces to return. Default to return all items.

        Yields:
            Union[Space, SpaceList]: yields Space one by one if chunk_size is not specified, else SpaceList objects.
        """
        return self._list_generator(
            list_cls=SpaceList,
            resource_cls=Space,
            method="GET",
            chunk_size=chunk_size,
            limit=limit,
        )

    def __iter__(self) -> Iterator[Space]:
        """Iterate over spaces

        Fetches spaces as they are iterated over, so you keep a limited number of spaces in memory.

        Yields:
            Space: yields Spaces one by one.
        """
        return cast(Iterator[Space], self())

    @overload
    def retrieve(self, space: str) -> Space | None:  # type: ignore[misc]
        ...

    @overload
    def retrieve(self, space: Sequence[str]) -> SpaceList:
        ...

    def retrieve(self, space: str | Sequence[str]) -> Space | SpaceList | None:
        """`Retrieve space by id. <https://docs.cognite.com/api/v1/#tag/Spaces/operation/bySpaceIdsSpaces>`_

        Args:
            space (str): Space ID

        Returns:
            Optional[Space]: Requested space or None if it does not exist.

        Examples:

                >>> from cognite.client import CogniteClient
                >>> c = CogniteClient()
                >>> res = c.data_modeling.spaces.retrieve(space='mySpace')

         Get multiple spaces by id:

                >>> from cognite.client import CogniteClient
                >>> c = CogniteClient()
                >>> res = c.data_modeling.spaces.retrieve(spaces=["MySpace", "MyAwesomeSpace", "MyOtherSpace"])

        """
        identifier = DataModelingIdentifierSequence.load_spaces(spaces=space)
        return self._retrieve_multiple(list_cls=SpaceList, resource_cls=Space, identifiers=identifier)

    def delete(self, space: str | Sequence[str]) -> list[str]:
        """`Delete one or more spaces <https://docs.cognite.com/api/v1/#tag/Spaces/operation/deleteSpacesV3>`_

        Args:
            space (str | Sequence[str]): ID or ID list ids of spaces.
        Returns:
            list[str]: The space(s) which has been deleted.
        Examples:

            Delete spaces by id::

                >>> from cognite.client import CogniteClient
                >>> c = CogniteClient()
                >>> c.data_modeling.spaces.delete(space=["mySpace", "myOtherSpace"])
        """
        deleted_spaces = cast(
            list,
            self._delete_multiple(
                identifiers=DataModelingIdentifierSequence.load_spaces(spaces=space), wrap_ids=True, returns_items=True
            ),
        )
        return [item["space"] for item in deleted_spaces]

    def list(
        self,
        limit: int = LIST_LIMIT_DEFAULT,
    ) -> SpaceList:
        """`List spaces <https://docs.cognite.com/api/v1/#tag/Spaces/operation/listSpacesV3>`_

        Args:
            limit (int, optional): Maximum number of spaces to return. Defaults to 25. Set to -1, float("inf") or None
                to return all items.

        Returns:
            SpaceList: List of requested spaces

        Examples:

            List spaces and filter on max start time::

                >>> from cognite.client import CogniteClient
                >>> c = CogniteClient()
                >>> space_list = c.data_modeling.spaces.list(limit=5)

            Iterate over spaces::

                >>> from cognite.client import CogniteClient
                >>> c = CogniteClient()
                >>> for space in c.data_modeling.spaces:
                ...     space # do something with the space

            Iterate over chunks of spaces to reduce memory load::

                >>> from cognite.client import CogniteClient
                >>> c = CogniteClient()
                >>> for space_list in c.data_modeling.spaces(chunk_size=2500):
                ...     space_list # do something with the spaces
        """
        return self._list(
            list_cls=SpaceList,
            resource_cls=Space,
            method="GET",
            limit=limit,
        )

    @overload
    def apply(self, space: Sequence[SpaceApply]) -> SpaceList:
        ...

    @overload
    def apply(self, space: SpaceApply) -> Space:
        ...

    def apply(self, space: SpaceApply | Sequence[SpaceApply]) -> Space | SpaceList:
        """`Create or patch one or more spaces. <https://docs.cognite.com/api/v1/#tag/Spaces/operation/ApplySpaces>`_

        Args:
            space (space: Space | Sequence[Space]): Space or spaces of spacesda to create or update.

        Returns:
            Space | SpaceList: Created space(s)

        Examples:

            Create new spaces:

                >>> from cognite.client import CogniteClient
                >>> from cognite.client.data_classes.data_modeling import SpaceApply
                >>> c = CogniteClient()
                >>> spaces = [SpaceApply(space="mySpace", description="My first space", name="My Space"),
                ... SpaceApply(space="myOtherSpace", description="My second space", name="My Other Space")]
                >>> res = c.data_modeling.spaces.apply(spaces)
        """
        return self._create_multiple(list_cls=SpaceList, resource_cls=Space, items=space)



from __future__ import annotations

from dataclasses import asdict, dataclass, field
from typing import ClassVar, Literal, Optional, Sequence, Tuple, Type, TypeVar, Union, cast

from cognite.client.utils._auxiliary import rename_and_exclude_keys
from cognite.client.utils._identifier import DataModelingIdentifier, DataModelingIdentifierSequence
from cognite.client.utils._text import convert_all_keys_recursive, convert_all_keys_to_snake_case


@dataclass
class DataModelingId:
    _type: ClassVar[str] = field(init=False)
    space: str
    external_id: str

    def as_tuple(self) -> tuple[str, str]:
        return self.space, self.external_id

    def dump(self, camel_case: bool = False, include_type: bool = True) -> dict[str, str]:
        output = asdict(self)
        if include_type:
            output["type"] = self._type
        return convert_all_keys_recursive(output, camel_case)

    @classmethod
    def load(cls: Type[T_DataModelingId], data: dict | T_DataModelingId | tuple[str, str]) -> T_DataModelingId:
        if isinstance(data, cls):
            return data
        elif isinstance(data, tuple):
            return cls(*data)
        elif isinstance(data, dict):
            return cls(**convert_all_keys_to_snake_case(rename_and_exclude_keys(data, exclude={"type"})))
        raise ValueError(f"Cannot load {data} into {cls}, invalid type={type(data)}")


T_DataModelingId = TypeVar("T_DataModelingId", bound=DataModelingId)


@dataclass
class VersionedDataModelingId:
    _type: ClassVar[str] = field(init=False)
    space: str
    external_id: str
    version: Optional[str] = None

    def as_tuple(self) -> tuple[str, str, Optional[str]]:
        return self.space, self.external_id, self.version

    def dump(self, camel_case: bool = False, include_type: bool = True) -> dict[str, str]:
        output = asdict(self)
        if include_type:
            output["type"] = self._type
        return convert_all_keys_recursive(output, camel_case)

    @classmethod
    def load(
        cls: Type[T_Versioned_DataModeling_Id], data: dict | T_Versioned_DataModeling_Id | tuple[str, str, str]
    ) -> T_Versioned_DataModeling_Id:
        if isinstance(data, cls):
            return data
        elif isinstance(data, tuple):
            return cls(*data)
        elif isinstance(data, dict):
            return cls(**convert_all_keys_to_snake_case(rename_and_exclude_keys(data, exclude={"type"})))
        raise ValueError(f"Cannot load {data} into {cls}, invalid type={type(data)}")


T_Versioned_DataModeling_Id = TypeVar("T_Versioned_DataModeling_Id", bound=VersionedDataModelingId)


@dataclass
class ContainerId(DataModelingId):
    _type = "container"


@dataclass
class ViewId(VersionedDataModelingId):
    _type = "view"


@dataclass
class DataModelId(VersionedDataModelingId):
    _type = "datamodel"


ContainerIdentifier = Union[ContainerId, Tuple[str, str]]
ViewIdentifier = Union[ViewId, Tuple[str, str], Tuple[str, str, str]]
DataModelIdentifier = Union[DataModelId, Tuple[str, str], Tuple[str, str, str]]

Id = Union[Tuple[str, str], Tuple[str, str, str], DataModelingId, VersionedDataModelingId]


def load_identifier(
    ids: Id | Sequence[Id], id_type: Literal["container", "view", "data_model"]
) -> DataModelingIdentifierSequence:
    is_sequence = isinstance(ids, Sequence) and not (isinstance(ids, tuple) and isinstance(ids[0], str))
    is_view_or_data_model = id_type in {"view", "data_model"}

    id_list = cast(Sequence, ids)
    if not is_sequence:
        id_list = [ids]

    return DataModelingIdentifierSequence(
        identifiers=[
            DataModelingIdentifier(
                *id_ if isinstance(id_, tuple) else (id_.space, id_.external_id, getattr(id_, "version", None))
            )
            for id_ in id_list
        ],
        is_singleton=not is_sequence and not is_view_or_data_model,
    )



from __future__ import annotations

from typing import Iterator, Sequence, cast, overload

from cognite.client._api_client import APIClient
from cognite.client._constants import DATA_MODELING_LIST_LIMIT_DEFAULT
from cognite.client.data_classes.data_modeling.data_models import (
    DataModel,
    DataModelApply,
    DataModelFilter,
    DataModelList,
)
from cognite.client.data_classes.data_modeling.ids import DataModelIdentifier, VersionedDataModelingId, load_identifier


class DataModelsAPI(APIClient):
    _RESOURCE_PATH = "/models/datamodels"

    @overload
    def __call__(
        self,
        chunk_size: None = None,
        limit: int = None,
        space: str | None = None,
        inline_views: bool = False,
        all_versions: bool = False,
        include_global: bool = False,
    ) -> Iterator[DataModel]:
        ...

    @overload
    def __call__(
        self,
        chunk_size: int,
        limit: int = None,
        space: str | None = None,
        inline_views: bool = False,
        all_versions: bool = False,
        include_global: bool = False,
    ) -> Iterator[DataModelList]:
        ...

    def __call__(
        self,
        chunk_size: int | None = None,
        limit: int = None,
        space: str | None = None,
        inline_views: bool = False,
        all_versions: bool = False,
        include_global: bool = False,
    ) -> Iterator[DataModel] | Iterator[DataModelList]:
        """Iterate over data model

        Fetches data model as they are iterated over, so you keep a limited number of data model in memory.

        Args:
            chunk_size (int, optional): Number of data model to return in each chunk. Defaults to yielding one data_model a time.
            limit (int, optional): Maximum number of data model to return. Default to return all items.
            space: (str | None): The space to query.
            inline_views (bool): Whether to expand the referenced views inline in the returned result.
            all_versions (bool): Whether to return all versions. If false, only the newest version is returned,
                                 which is determined based on the 'createdTime' field.
            include_global (bool): Whether to include global views.

        Yields:
            Union[DataModel, DataModelList]: yields DataModel one by one if chunk_size is not specified, else DataModelList objects.
        """
        filter = DataModelFilter(space, inline_views, all_versions, include_global)

        return self._list_generator(
            list_cls=DataModelList,
            resource_cls=DataModel,
            method="GET",
            chunk_size=chunk_size,
            limit=limit,
            filter=filter.dump(camel_case=True),
        )

    def __iter__(self) -> Iterator[DataModel]:
        """Iterate over data model

        Fetches data model as they are iterated over, so you keep a limited number of data model in memory.

        Yields:
            DataModel: yields DataModels one by one.
        """
        return cast(Iterator[DataModel], self())

    def retrieve(self, ids: DataModelIdentifier | Sequence[DataModelIdentifier]) -> DataModelList:
        """`Retrieve data_model(s) by id(s). <https://docs.cognite.com/api/v1/#tag/Data-models/operation/byExternalIdsDataModels>`_

        Args:
            ids (DataModelId | Sequence[DataModelId]): Data Model identifier(s).

        Returns:
            Optional[DataModel]: Requested data_model or None if it does not exist.

        Examples:

                >>> from cognite.client import CogniteClient
                >>> c = CogniteClient()
                >>> res = c.data_modeling.data_models.retrieve(("mySpace", "myDataModel", "v1"))

        """
        identifier = load_identifier(ids, "data_model")
        return self._retrieve_multiple(list_cls=DataModelList, resource_cls=DataModel, identifiers=identifier)

    def delete(self, ids: DataModelIdentifier | Sequence[DataModelIdentifier]) -> list[VersionedDataModelingId]:
        """`Delete one or more data model <https://docs.cognite.com/api/v1/#tag/Data-models/operation/deleteDataModels>`_

        Args:
            ids (DataModelId | Sequence[DataModelId]): Data Model identifier(s).
        Returns:
            The data_model(s) which has been deleted. None if nothing was deleted.
        Examples:

            Delete data model by id::

                >>> from cognite.client import CogniteClient
                >>> c = CogniteClient()
                >>> c.data_modeling.data_models.delete(("mySpace", "myDataModel", "v1"))
        """
        deleted_data_models = cast(
            list,
            self._delete_multiple(identifiers=load_identifier(ids, "data_model"), wrap_ids=True, returns_items=True),
        )
        return [
            VersionedDataModelingId(item["space"], item["externalId"], item["version"]) for item in deleted_data_models
        ]

    def list(
        self,
        limit: int = DATA_MODELING_LIST_LIMIT_DEFAULT,
        space: str | None = None,
        inline_views: bool = False,
        all_versions: bool = False,
        include_global: bool = False,
    ) -> DataModelList:
        """`List data models <https://docs.cognite.com/api/v1/#tag/Data-models/operation/listDataModels>`_

        Args:
            limit (int, optional): Maximum number of data model to return. Default to 10. Set to -1, float("inf") or None
                to return all items.
            space: (str | None): The space to query.
            inline_views (bool): Whether to expand the referenced views inline in the returned result.
            all_versions (bool): Whether to return all versions. If false, only the newest version is returned,
                                 which is determined based on the 'createdTime' field.
            include_global (bool): Whether to include global data models.

        Returns:
            DataModelList: List of requested data model

        Examples:

            List 5 data model:

                >>> from cognite.client import CogniteClient
                >>> c = CogniteClient()
                >>> data_model_list = c.data_modeling.data_models.list(limit=5)

            Iterate over data model:

                >>> from cognite.client import CogniteClient
                >>> c = CogniteClient()
                >>> for data_model in c.data_modeling.data_models:
                ...     data_model # do something with the data_model

            Iterate over chunks of data model to reduce memory load:

                >>> from cognite.client import CogniteClient
                >>> c = CogniteClient()
                >>> for data_model_list in c.data_modeling.data_models(chunk_size=10):
                ...     data_model_list # do something with the data model
        """
        filter = DataModelFilter(space, inline_views, all_versions, include_global)

        return self._list(
            list_cls=DataModelList,
            resource_cls=DataModel,
            method="GET",
            limit=limit,
            filter=filter.dump(camel_case=True),
        )

    @overload
    def apply(self, data_model: Sequence[DataModelApply]) -> DataModelList:
        ...

    @overload
    def apply(self, data_model: DataModelApply) -> DataModel:
        ...

    def apply(self, data_model: DataModelApply | Sequence[DataModelApply]) -> DataModel | DataModelList:
        """`Create or update one or more data model. <https://docs.cognite.com/api/v1/#tag/Data-models/operation/createDataModels>`_

        Args:
            data_model (data_model: DataModelApply | Sequence[DataModelApply]): DataModel or data model to create or update (upsert).

        Returns:
            DataModel | DataModelList: Created data_model(s)

        Examples:

            Create new data model::

                >>> from cognite.client import CogniteClient
                >>> import cognite.client.data_classes.data_modeling as models
                >>> c = CogniteClient()
                >>> data_models = [models.DataModel(space="mySpace",external_id="myDataModel",version="v1",is_global=,last_updated_time=),
                ... DataModel(space="mySpace",external_id="myOtherDataModel",version="v1",is_global=,last_updated_time=)]
                >>> res = c.data_modeling.data_models.create(data_models)
        """
        return self._create_multiple(list_cls=DataModelList, resource_cls=DataModel, items=data_model)
